#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE ~0
#endif

#ifdef QDOS
char _prog_name[] = "pcl2x 0.00";
#endif

typedef struct
{
    char key;
    void (*func) (FILE *);
} FTABLE;

typedef struct
{
    char key;
    FTABLE *table;
} KTABLE;


/* Table definitions */

static void DoAmpl (FILE *);

static FTABLE ampstab[] =
{
    {'l', DoAmpl},
    {0, NULL}
};

static void DoStarb (FILE *);
static void DoStarp (FILE *);
static void DoStarr (FILE *);
static void DoStart (FILE *);

static FTABLE startab[] =
{
    {'b', DoStarb},
    {'p', DoStarp},
    {'r', DoStarr},
    {'t', DoStart},
    {0, NULL}
};

static KTABLE keytab[] =
{
    {'*', startab},
    {'&', ampstab},
    {'E', NULL},
    {'z', NULL},
    {'X', NULL},
    {'Y', NULL},
    {'9', NULL},
    {0, NULL}
};


static HPDATA hpd;                  /* private HP data structure */
static CODERSTATE *code;            /* Coder structure */
                                    /* Most of this is PRIVATE */
                                    /* see g3encoder.h */
static char destdir[64] = DEFDIR;
static FAXDATA qd;                  /* QFAX data */
static char *usefile = NULL;

/*                     1         2         3          4 
 *           123 456789012345678901234567890123 45678901234567
 */
#define PCP "P4\n# pcl2pbm (c) Jonathan Hudson\n1600 00000000\n"
#define PCPLEN  47
#define PCPOFF  38

/* Open a PBM file (one per page). Zero counters */

int OpenPBM(char * fnam)
{
    int fd;
    fd = open(fnam, O_WRONLY|O_CREAT|O_TRUNC, 0666);
    write(fd, PCP, PCPLEN);
    hpd.faxline = hpd.lineno = 0;
    return fd;
}

/* Close PBM, get used lines and truncate trailing blank lines */

ClosePBM(void)
{
    char txt[10];
    int fsize;
    
    lseek(code->hd, PCPOFF, 0);
    sprintf(txt, "%8d", hpd.last);
    write(code->hd, txt, 8);
    fsize = PCPLEN + hpd.last * 200;
    ftruncate(code->hd, fsize);
    close(code->hd);
    code->hd = 0;    
}

/* Tell 'em as it is ... */

void Report (char *ctl,...)
{
    if (hpd.verbose)
    {
	va_list va;

	va_start (va, ctl);
	vprintf (ctl, va);
	va_end (va);
    }
}

/* reset HP state after page */

void Reset(void)
{
    bzero((char *) hpd.reflin, 300);    /* Should be void * !!! */
    hpd.faxline = 0;
    hpd.lineno = 0;    
}

/* HP mode 2 expander. */

int mode2expand (UCHAR * compressed, UCHAR * end, UCHAR * expanded)
{
    int nb = 0;
    short rc;
    UCHAR c;

    while (compressed < end)
    {
	rc = *compressed++;
	if (rc < 128)
	{
	    rc++;
	    memcpy (expanded, compressed, rc);
	    compressed += rc;
	}
	else if (rc > 128)
	{
	    rc = 257 - rc;
	    c = *compressed++;
	    memset (expanded, c, rc);
	}
	nb += rc;
	expanded += rc;
    }
    return nb;
}

/* Guess what ? */

void mode3expand (UCHAR * old, UCHAR * new, short nold, UCHAR * diff, short ndif)
{
    UCHAR *np = new;
    UCHAR val;
    short nbyte;
    int offset = 0;
    short xtra;

    if (ndif)
    {
	UCHAR *ep = diff + ndif;

	memcpy (new, old, nold);
	while (diff < ep)
	{

	    val = *diff++;
	    nbyte = 1 + (val >> 5);
	    offset = val & 0x1f;
	    if (offset == 31)
	    {
		do
		{
		    xtra = *diff++;
		    offset += xtra;
		}
		while (xtra == 255);
	    }
	    new = new + offset;
	    if (nbyte)
	    {
		memcpy (new, diff, nbyte);
		diff += nbyte;
		new += nbyte;
	    }
	}
    }
}

/* tog3
// This is the work of converting. We start 8 bytes (64 pixels) into the
// G3 line, we assume max of 2400 pixels at 300 dpi translated to 1600
// pixels at 200 dpi for fax. This centres the image. We take 2 out of 3
// bits and and 2 from 3 lines in high res (code->res == 1) or 1 from 3
// in low res (code->res =- 2).
*/
void tog3 (void)
{
    short bcp, bcf;
    UCHAR pixbuf[FBSIZE];
    UCHAR *hpbuf = hpd.datbuf; 
    UCHAR *g3buf = pixbuf;
    short black = 0;
    
    memset (g3buf, 0, FBSIZE);          /* Clear G3 data line */
    g3buf += 8;                         /* Start here for margins */

    for (bcp = bcf = 0; bcp < (hpd.datlen << 3);)
    {
	if ((bcp % 3))
	{
	    if (*hpbuf & code->mbits[bcp & 7])
	    {
		*g3buf |= code->mbits[bcf & 7];
	    }
	    bcf++;
	    if ((bcf & 7) == 0)
	    {
                if(*g3buf)
                {
                    black = 1;          /* This is a used (black) line */
                }
		g3buf++;
	    }
	}
	bcp++;
	if ((bcp & 7) == 0)
	{
	    hpbuf++;
	}
    }
    
    if(hpd.flag & HDR_FLG)          /* We need a BOP first */
    {
        if(hpd.pbm)
        {
            char *p;
            p = strrchr(hpd.fptr, SEP_CHAR);
            if(p)
            {
                sprintf(++p, "%d", hpd.page+1);
                code->hd = OpenPBM(hpd.fptr);
            }
        }
        else
        {
            FaxBeginPage();
        }
        hpd.flag &= ~HDR_FLG;
        hpd.flag |= RTC_FLG;
    }

    if(hpd.pbm)
    {
        if(code->hd == 0)
        {
            fputs("Write to ZERO\n", stderr);
        }
        write(code->hd, pixbuf+8, 200);
    }
    else
    {
        FaxCodeLine(pixbuf);           /* Encode G3 line */
    }
    hpd.faxline++;
    if(black)
    {
        hpd.last = hpd.faxline;     /* last black line on this page ? */
    }
}

/*  Encode a blank line (this is for HP vertical movements) */

void DoBlankLines (short cnt)
{
    UCHAR pixbuf[216];
    short i,j;

    bzero ((char *) pixbuf, FBSIZE);
    for(i = 0; i < cnt; i++)
    {
        if((j = (hpd.lineno % 3)))
        {
            if(code->res || (j == 1))
            {
                if(!hpd.pbm)
                {
                    FaxCodeLine (pixbuf);
                }
                else
                {
                    write(code->hd, pixbuf+8, 200);
                }
                hpd.faxline++;
            }
        }
        hpd.lineno++;
    }
}

/* Eat junk out of HP Data */

void Slurp (FILE * fp)
{
    int c;

    while ((c = fgetc (fp)), c != EOF && c >= '@' && c <= 'Z')
	/* NULL loop to eat junk */ ;

}

/* DoRTC
 * Handle form feed / page eject. Set new page flag and call EOP
 * code. For a fax, this is CoderEOP
 */

void DoRTC(void)
{
    if ((hpd.flag & RTC_FLG))
    {
        if(hpd.pbm)
        {
            ClosePBM();
        }
        else
        {
            FaxCodeEOP ();
	}
	hpd.flag &= ~RTC_FLG;
        hpd.flag |= HDR_FLG;
        hpd.lineno = 0;
        hpd.page++;
    }
}

/* parse a run of same type HP commands (till Upper Case letter) */

short ParseParams (FILE * fp, short *act, short *arg)
{
    char buf[128];		/* exceed and die */

    char *p, *ep;
    short done = 1;
    int c;

    *act = *arg = 0;

    ep = buf + sizeof (buf);
    for (p = buf; (c = fgetc (fp)) != EOF && p < ep;)
    {
	*p++ = c;

	if (isalpha (c))
	{
	    *act = c;
	    done = isupper (c);
	    *arg = strtol (buf, NULL, 10);
	    break;
	}
    }
    return done;
}


void DoAmpl (FILE * fp)
{
    short act, arg;
    short done;

    do
    {
	done = ParseParams (fp, &act, &arg);
	switch (act)
	{
	    case 'H':
	    case 'h':
		if (arg == 0)
		{
		    Report ("Eject page\n");
                    DoRTC();
		}
		break;
	}
    }
    while (!done);
}

void DoStarr (FILE * fp)
{
    short act, arg;
    short done;

    do
    {
	done = ParseParams (fp, &act, &arg);
	switch (act)
	{
	    case 'A':
		Report ("Start Graphics\n");
		break;
	    case 'B':
	    case 'C':
		if (arg == 0)
		{
		    Report ("End Graphics\n");
		}
		break;
	    case 'S':
	    case 's':
		Report ("Width %d\n", arg);
		break;
	}
    }
    while (!done);
}

void HandlePCLDat(void raw, int rawlen)
{
    Report ("Got %d bytes data\n", rawlen);
}


void DoStarb (FILE * fp)
{
    short act, arg;
    short done;
    char raw[512];
    
    do
    {
	done = ParseParams (fp, &act, &arg);
	switch (toupper (act))
	{
	    case 'W':
		fread (raw, 1, arg, fp);
                HandlePCLDat(raw);
		break;
	}
    }
    while (!done);
}

void DoStart (FILE * fp)
{
    short act, arg;
    short done;

    do
    {
	done = ParseParams (fp, &act, &arg);
	switch (toupper (act))
	{
	    case 'R':
		Report ("Resolution is %d dpi\n", arg);
		break;
	}
    }
    while (!done);
}

void DoStarp (FILE * fp)
{
    short act, arg;
    short done;
    int c;
    do
    {
        c = getc(fp); ungetc(c, fp); 
	done = ParseParams (fp, &act, &arg);
	switch (toupper (act))
	{
	    case 'X':
		Report ("Set X pos %d\n", arg);
		break;
	    case 'Y':
		Report ("Set Y pos %s%d\n", (c == '+') ? "+" : "", arg);
                if(c == '+')
                {
                    DoBlankLines(arg);
                }
		break;
	}
    }
    while (!done);
}


void DoTable (FILE * fp, FTABLE * p, char key)
{
    int c;

    c = fgetc (fp);
    for (; p->func; p++)
    {
	if (p->key == c)
	{
	    (p->func) (fp);
	    break;
	}
    }
    if (p->func == NULL)
    {
	Slurp (fp);
    }
}

/* We make specific files according to QFAX conventions */

#ifndef QDOS 
#ifdef __linux__
#define FilePart(p1) p1
#endif
char *MakeOutfile (char *infile)
{
    static char fpath[128];
    char *p;

    strcpy (fpath, destdir);
    if(usefile == NULL)
    {
        strcat (fpath, FilePart (infile));
        if ((p = strrchr (fpath, '.')) == NULL)
        {
	    p = fpath + strlen (fpath);
        }
    }
    else
    {
        strcat(fpath, usefile);
        p = fpath + strlen(fpath);
    }
    if(!hpd.pbm)
    {
        strcpy (p, qd.suffix);
    }
    else
    {
        sprintf(p, "pbm.%d", hpd.page+1);
    }
    return (hpd.fptr = fpath);
}
#else       /* QDOS */
char *MakeOutfile (char *filename)
{
    static char g3file[64];
    char *pp;
    char *ps = NULL;
    short use_fax = FALSE;

    if (usefile)
    {
	ps = usefile;
	use_fax = TRUE;
    }
    else
    {
	if (isdirdev (filename))
	{
	    ps = strchr (filename, '_');
	}
	if (ps == NULL)
	{
	    ps = filename;
	}
	else
	{
	    ps++;
	}
    }
    strcat (strcpy (g3file, destdir), ps);
    if(usefile == NULL)
    {
        if ((pp = strrchr (g3file, '_')) != NULL)
        {
	    *pp = 0;
        }
    }
    if(!hpd.pbm)
    {
        strcat(g3file, qd.suffix);
    }
    else
    {
        sprintf(g3file+strlen(g3file), "pbm_%d", hpd.page+1);
    }
    return (hpd.fptr = g3file);
}
#endif

int main (int ac, char **av)
{
    FILE *fp;
    int c;
    KTABLE *p;
    char fnam[64];
    short pack = 0; 
    short res = 0;
    
    /* 
     * Get QFAX parameters. We only use a few, like the temp field for
     * destination directory, and the default resolution
     */

    FaxParams (&qd);

    if (*qd.Temp)
    {
	strcpy (destdir, qd.Temp);
    }

    if(code = FaxInitCoder(qd.elsb[0]))
    {
    res = qd.res;

    while ((c = getopt (ac, av, "bvr:2u:x")) != -1)
    {
        switch(c)
        {
            case 'b':
                hpd.pbm = 1;
                break;
            case 'x':
#ifdef QDOS
            {
                extern char * _endmsg;
                _endmsg = NULL;
	    }
#endif
                break;
            case 'u':
                usefile = strdup(optarg);
                break;
            case 'v':
                hpd.verbose = 1;
                break;
            case '2':
                pack = 1;
                break;
            case 'r':
                res = atoi(optarg);
                break;
        }
    }
    if(av[optind])
    {
        strcpy(fnam, av[optind]);
        fp = fopen (fnam, "r");
    }
    else
    {
        fp = stdin;
#ifdef _DCC
        strcpy(fnam, "pcl2hp.hp");
#else
        strcpy(fnam, "ram1_pcl2hp_hp");
#endif
    }
    if (fp)
    {
        char *q;
        hpd.reflen = 300;
        
        hpd.flag = RTC_FLG;
        hpd.datbuf = hpd.datlin;
        q = MakeOutfile (fnam);        
        if(hpd.pbm)
        {
            code->hd = OpenPBM(q);
            code->pix = 1728;
            code->res = 1;
	}
	else
	{
            code->hd = FaxOpenFile (q, res, pack, FPSIZE);
            FaxHeader (0, 0, 0, NULL, 0);    /* c++ please */
            FaxBeginPage();            
        }    
        if(code->hd > 0)
	{
	    while ((c = fgetc (fp)) != EOF)
	    {
		if (c == 27)
		{
		    c = fgetc (fp);
		    for (p = keytab; p->key; p++)
		    {
			if (p->key == c)
			{
			    if (p->table)
			    {
				DoTable (fp, p->table, c);
				break;
			    }
			    else
			    {
			        if(c == 'E')
			        {
			            Reset();
			            break;
			        }
			    }
			}
		    }
		}
		else if (c == 12)
		{
		    Report ("Form feed\n");
                    DoRTC();
		}
	    }
            if ((hpd.flag & RTC_FLG))
            {
                fputs("assume EOF !!\n", stderr);
                DoRTC();
            }
	    if(!hpd.pbm)
	    {
                FaxFlushBuf(((code->bitnr+7) >> 3));
                FaxHeader(hpd.page, WIDTH_A4, UNLIMITED, qd.LocalId, 0);
            }
            if(code->hd)
            {
                close(code->hd);
            }
	}
    }
        FaxFreeCoder(code);
    }
    else
    {
        fputs("No memory\n", stderr);
    }
    return 0;
}
