# include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifndef QDOS
# include <dos/stdio.h>
# include <dos/dos.h>
# include <exec/exec.h>
# include <clib/dos_protos.h>
# include <clib/exec_protos.h>
# define NOT_FOUND 6 
# define QFAX_PHONE "qfax.phone"
#else
# define Open fopen
# define MODE_OLDFILE "r"
# define MODE_NEWFILE "w"
# define Close fclose
# define FGets(p1, p2, p3) fgets(p2, p3, p1)
# define FPutC(p1, p2) fputc(p2, p1)
# define PutStr(p1) fputs(p1, stderr)
# define BPTR FILE * 
# define Printf printf
# define FPrintf fprintf
# define NOT_FOUND (-7)
# define QFAX_PHONE "qfax_phone"
#endif
 
#include "qfaxparse.h"


static char *strim (char *s)
{
    char *p;
    short n;

    n = strlen(s);
    p = s + n;
    while(*--p <= ' ' && n--)
    {
        *p = '\0';
    }
    return s;
}

static void  pAlias (PFAXDATA p, char *s, int nx)
{
    char *p1,*p2;
    short n;
    ALIAS *pn;

    pn = (ALIAS *) calloc (sizeof(ALIAS), 1);

    if(pn)
    {
        for(p1 = s, n = 0; ; n++)
        {
            int k;
            char c, *p3;
            p2 = strchr(p1, '/');
            if(p2)
            {
                k = (p2-p1);
            }
            else
            {
                k = strlen(p1);
            }
            p3 = (p1 + k);
            c = *p3;
            *p3 = 0;
            switch(n)
            {
                case 0:
                    strncpy (pn->phone, p1, 39);
                    break;
                case 1:
                    strncpy (pn->alias, p1, 39);
                    break;
                case 2:
                    strncpy (pn->abbrev, p1, 15);
                    break;                
                case 3:
                    strncpy (pn->dcode, p1, 39);
                    break;
                case 4:
                    pn->flag = strtol(p1, NULL, 0);
                    break;
                case 5:
                    strncpy (pn->dcode, p1, 39);
                    pn->flag |= 0x8000;
                    break;
            }
            if(!(*p3 = c))
            {
                break;
            }
            p1 = p2+1;
        }
    }
    pn->next = p->alias;
    p->alias = pn;
}
void FreeAlias(PFAXDATA p)
{
    ALIAS *pl, *dead;

    for(pl = p->alias; pl;)
    {
        dead = pl;
        pl = pl->next;
        free (dead);
    }
    p->alias = NULL;
}


static PARSELIST pl[] = {{"ALIAS",   pAlias,     0}, NULL};

void FaxParams (PFAXDATA p)
{
    BPTR fp;
    char *pf;
    
#ifndef QDOS
    char pname[64];

    if(GetVar("QFAX_DATA", (pf = pname), sizeof(pname), 0) <= 0)
    {
        pf = QFAX_FILE;
    }
#else
    if((pf = getenv("QFAX_DATA")) == NULL)
    {
#ifndef _DCC
        pf = qfd.name;
#else            // pretend to be QDOS :-(
        pf = "qfax_dat";
#endif
    }
#endif

    if(fp = Open(pf, MODE_OLDFILE))
    {
        char text [128];

        while(FGets(fp, text, sizeof(text)))
        {
            char *s;
            PPARSELIST ppl;
            char *ptr;

            ptr = (char *)p;
            strim(text);
            
            for(ppl = pl; ppl->id; ppl++)
            {
                if(strncasecmp(text, ppl->id, strlen(ppl->id)) == 0)
                {
                    int l;
                    if((s = strchr(text, '=')))
                    {
                        s++;
                        l = strspn(s, " \t");
                    }
                    else if(ppl->flag)
                    {
                        s = text+strlen(text);
                        l = 0;
                    }
                    if(s)
                    {
                        (ppl->func)((ptr + ppl->offset), s + l, ppl->mval);
                        break;
                    }
                }
            }
        }
        Close(fp);
    }
    else
    {
        p->nodata = 1;
    }
}


int main(int ac, char **av)
{
static FAXDATA ff = {0};
    
    FaxParams(&ff);
    if(ff.alias)
    {
        BPTR fd = 0;
        ALIAS *pl;

        fd = Open(QFAX_PHONE, MODE_NEWFILE);
        for(pl = ff.alias; pl; pl = pl->next)
        {
            Printf("Alias: %s %s %s %lx (%s)\n",
            pl->alias, pl->dcode, pl->phone, pl->flag, pl->abbrev);
            if(fd)
            {
                if(*pl->alias)
                {
                    FPrintf(fd,"NAME=%s, ", pl->alias);
                }
                
                if(*pl->abbrev)
                {
                    FPrintf(fd,"ALIAS=%s, ", pl->abbrev);
                }
                if(*pl->phone)
                {
                    FPrintf(fd,"TSI=%s, ", pl->phone);
                }
                if(*pl->dcode)
                {
                    if(pl->flag & 0x8000)
                    {
                        FPrintf(fd,"FULL=%s", pl->dcode);
                    }
                    else
                    {
                        FPrintf(fd,"DCODE=%s", pl->dcode);
                    }
		}
	        FPutC(fd, '\n');	
            }
	}
        FreeAlias(&ff);
        if(fd)
        {
            Close(fd);
        }
    }
    return 0;
}

