#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include "gdevtiff.h"
#include "qheader.h"
#include "/lib/qfaxparse.h"
#include "/lib/qheader.h"

#ifdef _DCC
#include <getopt.h>
#endif

#include "pqf.h"

#define notdef 1

static VIEWPORT vp = {0};
PVIEWPORT pv = &vp;

typedef struct
{
    short flag;
    short xpix;
    short ypix;
    short llen;
    short mode;
} QHEAD, *PQHEAD;


#define QL_SCREEN   0
#define BIT_MAP     1
#define DO_GIF      2
#define TIFF_F      3
#define PCX         4
#define PBM         5
#define PBMRAW      6

extern int decode_fax(void);
extern void GiveItBack(void);
extern int DoGif(FILE *, char *, short);
extern int dopcx (FILE * file, char * map, short height);
extern int dopbm (char, FILE * file, char * map, short height);

#define MAX_FAX (2240*216)

#define PROTO_ARGS(X)	X

#define X_DPI 204
#define Y_DPI 196

static short qp = 1;

typedef struct {
    FILE*	fp;
    long	prevdir;	/* file offset of previous directory offset */
    long	diroff;		/* file offset of next write */
    int		bigendian;	/* 1 if machine is big-endian, 0 otherwise */
    int		fax_byte;
    int		fax_weight;
} TIFFOUT;

/* NB: this array is sorted by tag number (assumed below) */
typedef struct {
    TIFFDirEntry	subfiletype;        //0
    TIFFDirEntry	imagewidth;         //1
    TIFFDirEntry	imagelength;        //2
    TIFFDirEntry	bitspersample;      //3
    TIFFDirEntry	compression;        //4
    TIFFDirEntry	photometric;        //5
    TIFFDirEntry	fillorder;          //6
#ifdef notdef
    TIFFDirEntry	documentname;
#endif
    TIFFDirEntry	stripoffsets;       //7
    TIFFDirEntry	orientation;        //8
    TIFFDirEntry	samplesperpixel;    //9
    TIFFDirEntry	rowsperstrip;       //10
    TIFFDirEntry	stripbytecounts;    //11
    TIFFDirEntry	xresolution;        //12
    TIFFDirEntry	yresolution;        //13
    TIFFDirEntry	planarconfig;       //14
    TIFFDirEntry	group3options;      //15
    TIFFDirEntry	resolutionunit;     //16
#ifdef notdef
    TIFFDirEntry	software;           
#endif
    TIFFDirEntry	cleanfaxdata;       //17
    unsigned long	diroff;			/* offset to next directory */
    unsigned long	xresValue[2];		/* xresolution indirect value */
    unsigned long	yresValue[2];		/* yresolution indirect value */
} TIFFDirectory;
static TIFFDirectory dirTemplate = {
    { TIFFTAG_SUBFILETYPE,	TIFF_LONG,  1, FILETYPE_PAGE },
    { TIFFTAG_IMAGEWIDTH,	TIFF_LONG,  1 },
    { TIFFTAG_IMAGELENGTH,	TIFF_LONG,  1 },
    { TIFFTAG_BITSPERSAMPLE,	TIFF_SHORT, 1, 1 },
    { TIFFTAG_COMPRESSION,	TIFF_SHORT, 1, COMPRESSION_CCITTFAX3 },
    { TIFFTAG_PHOTOMETRIC,	TIFF_SHORT, 1, PHOTOMETRIC_MINISWHITE },
    { TIFFTAG_FILLORDER,	TIFF_SHORT, 1, FILLORDER_MSB2LSB },
#ifdef notdef
    { TIFFTAG_DOCUMENTNAME,	TIFF_ASCII, 1 },
#endif
    { TIFFTAG_STRIPOFFSETS,	TIFF_LONG,  1 },
    { TIFFTAG_ORIENTATION,	TIFF_SHORT, 1, ORIENTATION_TOPLEFT },
    { TIFFTAG_SAMPLESPERPIXEL,	TIFF_SHORT, 1, 1 },
    { TIFFTAG_ROWSPERSTRIP,	TIFF_LONG,  1, -1L },
    { TIFFTAG_STRIPBYTECOUNTS,	TIFF_LONG,  1, 1 },
    { TIFFTAG_XRESOLUTION,	TIFF_RATIONAL, 1 },
    { TIFFTAG_YRESOLUTION,	TIFF_RATIONAL, 1 },
    { TIFFTAG_PLANARCONFIG,	TIFF_SHORT, 1, PLANARCONFIG_CONTIG },
    { TIFFTAG_GROUP3OPTIONS,	TIFF_LONG,  1 },
    { TIFFTAG_RESOLUTIONUNIT,	TIFF_SHORT, 1, RESUNIT_INCH },
#ifdef notdef
    { TIFFTAG_SOFTWARE,		TIFF_ASCII, 1 },
#endif
    { TIFFTAG_CLEANFAXDATA,	TIFF_SHORT, 1, CLEANFAXDATA_CLEAN },
    0, { 0, 1 }, { 0, 1 },
};
#define	OFFSET(x)	&(((TIFFDirectory*)NULL)->x)
#define	NTAGS		((int) OFFSET(diroff) / sizeof (TIFFDirEntry))
/* correct tag values on bigendian machines */

static unsigned char eol_lead[256] = {0};
static unsigned char eol_tail[256] = {0};
static unsigned char swap_table[256] = {0};

void Report (char *ctl,...)
{
    va_list va;
    va_start (va, ctl);
    vprintf(ctl, va);
    va_end (va);
}

static void faxout_fixuptags(TIFFDirEntry* dp, int n)
{
    while (n-- > 0) {
	if (dp->tdir_type == TIFF_SHORT || dp->tdir_type == TIFF_SSHORT)
	    dp->tdir_offset <<= 16;
	else if (dp->tdir_type == TIFF_BYTE || dp->tdir_type == TIFF_SBYTE)
	    dp->tdir_offset <<= 24;
	dp++;
    }
}

static void faxout_open_fp(FILE *fp, register TIFFOUT* faxp)
{
    faxp->fp = fp;
    faxp->diroff = 0L;
    faxp->prevdir = 0L;
    faxp->bigendian = 1;
    faxp->fax_byte = 0;
    faxp->fax_weight = 0x80;
}

int faxout_begin_page(TIFFOUT *faxp, short res)
{
    short dircount;
    TIFFDirectory dir;
    int m,n;
#ifdef notdef
    static char *docnam = "QFAX TIFF/F Document";
    static char *swname = "qfu (c) Jonathan Hudson";
#endif
    /*
     * Writing the header is delayed to here because the
     * FILE* is not setup when faxout_open is called.
     */
    if (faxp->diroff == 0) {
	TIFFHeader h;
	h.tiff_magic = (faxp->bigendian ? TIFF_BIGENDIAN : TIFF_LITTLEENDIAN);
	h.tiff_version = TIFF_VERSION;
	h.tiff_diroff = sizeof (TIFFHeader);
	fwrite((char*) &h, sizeof (h), 1, faxp->fp);
	faxp->diroff = sizeof (TIFFHeader);	/* where next directory goes */
    } else {
	/* patch pointer to this directory */
	fseek(faxp->fp, faxp->prevdir, 0);
	fwrite((char*)&faxp->diroff, sizeof (faxp->diroff), 1, faxp->fp);
    }
    fseek(faxp->fp, faxp->diroff, 0);
    /* write count of tags in directory */
    dircount = NTAGS;
    fwrite((char*)&dircount, sizeof (dircount), 1, faxp->fp);
    faxp->diroff += sizeof (dircount);

    /* fill in directory tags and write them */
    memcpy(&dir, &dirTemplate, sizeof (dirTemplate));
    dir.imagewidth.tdir_offset = 1728;

    dir.imagelength.tdir_offset = (res) ? 2240 : 1120;
    dir.compression.tdir_offset = COMPRESSION_CCITTFAX3;
    dir.stripoffsets.tdir_offset = faxp->diroff + sizeof (TIFFDirectory);
#ifdef notdef
    dir.documentname.tdir_offset = dir.stripoffsets.tdir_offset; 
    dir.documentname.tdir_count = strlen(docnam);
    dir.software.tdir_offset = dir.documentname.tdir_offset + dir.documentname.tdir_count + 1;
    dir.software.tdir_count = strlen(swname);
    dir.stripoffsets.tdir_offset =  dir.software.tdir_offset + dir.software.tdir_count + 1;
#endif
    m = (int) OFFSET(xresValue);
    dir.xresolution.tdir_offset = faxp->diroff + m;
    n = (int) OFFSET(yresValue);
    dir.yresolution.tdir_offset = faxp->diroff + n;
    dir.group3options.tdir_offset = 0;
    dir.xresValue[0] = X_DPI;
    dir.yresValue[0] = (res) ? Y_DPI : Y_DPI/2;
    if (faxp->bigendian)
    {
	faxout_fixuptags(&dir.subfiletype, NTAGS);
    }
    fwrite((char*)&dir, sizeof (dir), 1, faxp->fp);
#ifdef notdef
    fwrite(docnam, 1, strlen(docnam) + 1, faxp->fp);
    fwrite(swname, 1, strlen(swname) + 1, faxp->fp);
#endif

    return (0);
}

static int faxout_end_page(TIFFOUT *faxp, int nlines)
{
    long diroff, cc;

    diroff = faxp->diroff;
    faxp->prevdir = faxp->diroff + (int) OFFSET(diroff);
    faxp->diroff = ftell(faxp->fp);
    /* patch strip byte counts value */
    cc = faxp->diroff - (diroff + sizeof (TIFFDirectory)) - 8;
    fseek(faxp->fp, diroff + (int) OFFSET(stripbytecounts.tdir_offset), 0);
    fwrite(&cc, sizeof (cc), 1, faxp->fp);
    fseek(faxp->fp, diroff + (int) OFFSET(imagelength.tdir_offset), 0);
    fwrite(&nlines, sizeof (int), 1, faxp->fp);
    return (0);
}

static void BuildTables (void)
{
    int i, j, l0, t0;
    
    for (i = 0; i < 256; i++)
    {
	swap_table[i] =
	    (i & 1 ? 128 : 0) | (i & 2 ? 64 : 0) | (i & 4 ? 32 : 0) | (i & 8 ? 16 : 0) |
	    (i & 16 ? 8 : 0) | (i & 32 ? 4 : 0) | (i & 64 ? 2 : 0) | (i & 128 ? 1 : 0);

	l0 = t0 = 0;
	for (j = 0; j < 8; j++)
	{
	    if (!(i & (0x80 >> j)))
	    {
		t0++;
	    }
	    else
	    {
		break;
	    }
	}
	for (j = 0; j < 8; j++)
	{
	    if (!(i & (1 << j)))
	    {
		l0++;
	    }
	    else
	    {
		break;
	    }
	}
	eol_lead[i] = l0;
	eol_tail[i] = t0;
    }
}

MoveQfax(FILE *f1, FILE *f2)
{

    int i,j;
    unsigned char buf[512];
    int zeros = 0, eolns = 0, eleven0 = 0;
    int c;
    int nl = 0;
    
    BuildTables();

   while((j = fread(buf, 1, 512, f1)) > 0)
    {
        for(i = 0; i < j  ; i++)
        {
            fputc(buf[i], f2);
            c = swap_table[buf[i]];
	    if (zeros + eol_lead[c] >= 11)
		eleven0 = 1;
	    else
	    {
		if (c)
		{
		    zeros = eol_tail[c];
		    eolns = 0;
		}
		else
		    zeros += eol_tail[c];
	    }
	    if ((eleven0) && c)
	    {
		eleven0 = 0;
		eolns++;
		nl++;
		zeros = eol_tail[c];
		if (eolns > 5)
		{
		    eolns = 0;
                    break;
		}
	    }
        }
    }
    nl -= 6;            // Magic
    
    return nl;
}

QLWrite(FILE *fp, unsigned char *p, PQHEAD pq)
{
    unsigned char r[480], *s, *q;
    int i;
    
    s = r;
    *s = 0;
    q = p;

    memset (r, '\0', sizeof(r));
    
    for(i = 0; i < pq->llen/2;i++)
    {
        *q = ~*q;
        *s++ = *q;
        *s++ = *q;
        q++; 
    }
    fwrite(r, pq->llen, 1, fp);
}

WriteQL(FILE *fp, char *map, short nl)
{
    QHEAD qh;
    unsigned char *d;
    short last,j,actual;
            
    qh.flag = 0x4afc;
    qh.xpix = (qp) ? 1728 : 512;
    qh.ypix = 0;
    qh.llen = 2*((qh.xpix+7)/8);
    qh.mode = 0;
    if(qp)
    {
        fwrite(&qh, sizeof(qh), 1, fp);
    }    

    last = (qp) ? nl : (nl > 342) ? 342 : nl;
 
    d = (unsigned char *) map;
    actual = 0;
    
    for(j = 0; j < last; j++)
    {
        d += 216;
        if ((j % 4))
        {
            QLWrite(fp, d, &qh);
            actual++;
        }
    }
    if(qp)
    {
        fseek(fp, 0, 0);
        qh.ypix = actual;
        fwrite(&qh, sizeof(qh), 1, fp);
    }        
    return 0;
}

DoTiff_F(FILE *fp1, char *p, PQFAXHEADER h)
{
    FILE *fp2;
    TIFFOUT tiff;
    int nlines;
    int nc;

    if(fp2 = fopen(p, "w"))
    {
        faxout_open_fp(fp2, &tiff);
        faxout_begin_page(&tiff, h->resolution);
        nlines = MoveQfax(fp1, fp2);
        faxout_end_page(&tiff, nlines);
        fseek(fp2, 0L, SEEK_END);
        nc = ftell(fp2);
        ftruncate(fileno(fp2), nc-8);
        fclose(fp2);
    }
    else
    {
        perror("Output file");
    }
}

DoMap (char *p, char *q, PQFAXHEADER h, short flag)
{
    FILE *fp2;
    int nc;

    if(fp2 = fopen(q, "w"))
    {
        if((pv->scan = (char *) malloc(MAX_FAX)) != NULL)
        {
            memset(pv->scan, '\0', MAX_FAX);
            strcpy(pv->FileName, p);
            if(decode_fax() == 0)
            {
                switch(flag)
                {
                    case BIT_MAP:
                        fwrite(pv->scan, 1, (216*pv->xline), fp2);
                        break;
                    case DO_GIF:
                        DoGif(fp2, pv->scan, pv->xline);
                        break;
                    case QL_SCREEN:
                        WriteQL(fp2, pv->scan, pv->xline);
                        break;
                    case PCX:
                        dopcx (fp2, pv->scan, pv->xline);
                        break;
                    case PBM:
                        dopbm ('1', fp2, pv->scan, pv->xline);
                        break;
                    case PBMRAW:
                        dopbm ('4', fp2, pv->scan, pv->xline);
                        break;
                }
            }
#ifdef _DCC
            GiveItBack();
#endif
            free (pv->scan);
            fclose(fp2);
        }
    }
    else
    {
        perror("Output file");
    }
}


main(int ac, char **av)
{
    FILE *fp1;
    QFAXHEADER h;
    short flag;
    int c;
    char *fn1, *fn2;
    short mflag = 0;
    
    while((c = getopt(ac, av, "f:htm")) != EOF)
    {
        switch(c)
        {
            case 'm':
                mflag = 1;
                break;
            case 'f':
                flag = atoi(optarg);
                break;
            case 't':
                qp = 0;
                break;
            case 'h':
            case '?':
                puts("qfu -f n infile outfile");
                puts("  n = (0,1,2,3,4,5) Qpaint,Bitmap,GIF,TIFF/F.PCX,PBM,PBMraw");
                break;
        }
    }
  
#ifdef QDOS
    if(mflag == 0)
    {
        extern char *_endmsg;
        _endmsg = NULL;
    }
#endif

    fn1 = *(av + optind);
    fn2 = *(av + optind + 1);

    if(!fn1 || !fn2)
    {
        puts("Invalid file names");
        exit(0);
    }
    
    if(fp1 = fopen(fn1, "r"))
    {
        fread(&h, sizeof(h), 1, fp1);
        if (h.title == QFAXTITLE)
        {
            if(flag == TIFF_F)
            {
                DoTiff_F(fp1, fn2, &h);
            }
            else
            {
                fclose(fp1);
                fp1 = NULL;
                DoMap (fn1, fn2, &h, flag);
            }
        }
        else
        {
            puts("Not QFAX");
        }
        if(fp1)
            fclose(fp1);
    }
    else
    {
        perror("Input file");
    }
    return 0;
}
