/********************************************************
 ** 2tif - SMS2/QDOS to GIF (c) Jonathan Hudson 1994
 **
 ********************************************************/

/*======================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <tiffio.h>

#define OEXTRA ,0

#define private static

typedef unsigned char byte;
#ifndef __unix
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned long ulong;
#endif

typedef struct
{
    short flag;
    short xpix;
    short ypix;
    short llen;
    short zero;
}
QHEAD, *PQHEAD;

typedef struct
{
    int fd;
    char *op;
    QHEAD qh;
    short comp;
    short qlar;
}
QLS, *PQLS;


static TIFF *OpenTIFF(int w, int h, char *fname, int comp)
{
  static ushort r[4] = {0, 0xffff, 0,      0xffff};
  static ushort g[4] = {0, 0,      0xffff, 0xffff};
  static ushort b[4] = {0, 0,      0,      0xffff};
  TIFF *tif;

  tif = TIFFOpen(fname, "w");
  if (!tif) return 0;

  TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, w);
  TIFFSetField(tif, TIFFTAG_IMAGELENGTH, h);
  TIFFSetField(tif, TIFFTAG_COMPRESSION, comp);
  TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 1);
  TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
  TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, h);
  TIFFSetField(tif, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
  TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 2);
  TIFFSetField(tif, TIFFTAG_COLORMAP, r, g, b);
  TIFFSetField(tif, TIFFTAG_PHOTOMETRIC,    PHOTOMETRIC_PALETTE);
  TIFFSetField(tif, TIFFTAG_SOFTWARE, "2tif - (c) Jonathan Hudson 1994");
  return tif;
}

static void MakeChunkyPix (byte * cpix, byte * qpix, short xpix)
{
    byte *q, *p;
    ushort k;
    ushort j;
    ushort l;
    ushort gb, rb, n;
    short llen, ns, m;
	
    llen = ((xpix+3) & ~3) >> 2;

    ns = (xpix & 3);
    
    p = cpix;
    q = qpix;
    memset (cpix, '\0', llen);
    l = k = 7;

    for (m = 0; m < xpix; m++)
    {
	j = 1 << k;
	gb = (*q & j);
	if (gb)
	{
	    *p |= (1 << l);
	}

	l--;

	rb = *(q + 1) & j;
	if (rb)
	{
	    *p |= (1 << l);
	}

	if (l == 0)
	{
	    p++;
	    l = 7;
	}
	else
	{
	    l--;
	}

	if (k == 0)
	{
	    k = 7;
	    q += 2;
	}
	else
	{
	    k--;
	}
    }
}

/* Write a page to a file in TIF format. */

int DoTIF (PQLS pq)
{
    byte *qldat = (byte *) malloc (pq->qh.llen);
    byte *pic;
    short k,y,j;
    short height;
    TIFF *tif;
    
    height = pq->qh.ypix;

    if(pq->qlar)
    {
        height = height * pq->qlar / (pq->qlar - 1);
    }
    pic =  malloc(pq->qh.llen * 2);

    tif= OpenTIFF(pq->qh.xpix, height, pq->op, pq->comp);

    fprintf(stderr, " (%d x %d)\n", pq->qh.xpix, height);
    
    for (k = y = 0; k < height;k++)
    {
        if(pq->qlar == 0 || (((k+1) % pq->qlar) && y < pq->qh.ypix))
        {
            read (pq->fd, qldat, pq->qh.llen);
            y++;
        }
	MakeChunkyPix (pic, qldat, pq->qh.xpix);
        TIFFWriteScanline(tif, pic, k, 0);
    }

    free ((char *) pic);
    free ((char *) qldat);

    TIFFClose(tif);
    return 0;
}


main (int ac, char **av)
{
    QLS qs;
    struct stat s;
    char c;
    
    char *ip = NULL;

    qs.op = NULL;
    qs.comp = COMPRESSION_LZW;
    qs.qlar = 4;
    
    while(++av,--ac)
    {
        if(**av == '-')
        {
            switch(*++(*av))
            {
#ifdef QDOS
                case 'q':
                {
                extern char * _endmsg;
                _endmsg = NULL;
                }
                break;
#endif

    	        case 'i':
                    qs.qlar = 0;
                    break;
    	        case 'a':
                    qs.qlar = 3;
                    break;
                case 'c':
                    c = *++(*av);
                    if(c == 'b' || c == 'p')
                    {
                        qs.comp = COMPRESSION_PACKBITS;
                    }
                    else if (c == '0' || c == '\0')
                    {
                        qs.comp = COMPRESSION_NONE;
		    }
                    break;
            }
        }
	else
	{
            if(ip == NULL)
            {
                ip = *av;
            }
            else if(qs.op == NULL)
            {
                qs.op = *av;
            }
        }
    }

    if ((qs.fd = open (ip, O_RDONLY, 0)) > 0)
    {
#ifndef QDOS
#define leus(v) ((v) >> 8) + ((v) << 8)
        {
            QHEAD q1;
            read (qs.fd, &q1, sizeof (QHEAD));
            qs.qh.flag = leus(q1.flag); 
            qs.qh.xpix = leus(q1.xpix); 
            qs.qh.ypix = leus(q1.ypix); 
            qs.qh.llen = leus(q1.llen); 
        }
#else
	read (qs.fd, &qs.qh, sizeof (QHEAD));
#endif
	if (qs.qh.flag != 0x4afc)
	{
	    fstat (qs.fd, &s);
	    qs.qh.xpix = 512;
	    qs.qh.llen = 128;
	    qs.qh.ypix = (s.st_size + 127) / 128;
	    lseek (qs.fd, 0, 0);
	}

        fprintf(stderr, "TIFFing %s to %s (%d x %d) -> ", ip, qs.op,
            qs.qh.xpix, qs.qh.ypix);
        fflush(stderr);            
        DoTIF (&qs);
	close (qs.fd);
    }
    else
    {
	perror ("input");
    }
}

