#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "qfax.h"

typedef struct
{
    int vr;
    int br;
    int wd;
    int ln;
    int df;
    int ec;
    int bf;
    int st;
} T30PARAMS, *PT30PARAMS;

extern int hsc, ppr, good, poll, ptsi, fline;
extern char tsi[];
extern short c20;

extern T30PARAMS x30p, t30p;
 
static int parse_t30_response(char *reply, T30PARAMS *p, void *x)
{
    char *ptr;
    int rep = -1;
    
    if ((ptr = strchr(reply, ':')) != NULL)
    {
	rep = (sscanf(ptr+1, "%d,%d,%d,%d,%d,%d,%d,%d",
	       &p->vr, &p->br, &p->wd, &p->ln,
		      &p->df, &p->ec, &p->bf, &p->st) != 8);
    }
    return (rep);
}

static int parse_int_response (char *reply, int *value, void *x)
{
    char *p;
    int rep = -1;
    int mx = 0;
    
    if (NULL != (p = strchr(reply, ':')))
    {
        int nc = p - reply + 1;
	if(x)
	{
	    *(int *)x = 0;
	}
	rep = (sscanf(reply+nc, "%d,%d", value, &mx) <  1);

	if(rep == 0 && x)
        {
            *(int *)x = mx;
        }
	if (rep == 0 && value == &ppr)
	{
	    good = (ppr & 7) & 1;
	}
    }
    return rep;
}
static int parse_flag(char *reply, int *flag, void *x)
{
    *flag = 1;
    return 0;
}
static int parse_id_response (char *reply, char *str, int *extra)
{
    char *fquote, *lquote;
    
    /* There are no quotes (") in ZyXEL response */
    if (strchr (reply, '"') == NULL)
    {
	char *s;
	if ((fquote = strchr (reply, ':')) == NULL)
	    return (-1);
	fquote++;
	while (isspace (*fquote))
	    fquote++;
	strcpy (str, fquote);
	for (s = str + strlen (str); s >= str; s--)
	{
	    if (!isgraph(*s))
	    {
		*s = '\0';
	    }
	    else
	    {
	        break;
	    }
	}
	if (strlen (str) == 0)
	{
	    strcpy (str, "Unknown");
	}
    }
    else
    {
	if ((fquote = strchr (reply, '"')) == NULL)
	{
	    return (-1);
        }
	if ((lquote = strrchr (reply, '"')) == NULL)
	{
	    return (-1);
	}
	if (fquote == lquote)
	{
	    return (-1);
	}
	*lquote = '\0';
	strcpy (str, fquote + 1);
    }
    if(extra)
    {
        *extra = 1;
    }
    return 0;
}

#define IFCVV int(*)(char *, void *, void *)

typedef struct pswitch {
    char *string;
    int (*func)(char *, void *, void *);
    void *ptr;
    void *extra;
} PSW;

static PSW psw0[] = {
	{"+FHNG", (IFCVV) parse_int_response, &hsc, NULL},
	{"+FPTS", (IFCVV) parse_int_response, &ppr, &fline},
	{"+FDCS", (IFCVV) parse_t30_response, &t30p},
	{"+FDIS", (IFCVV) parse_t30_response, &x30p},
	{"+FDTC", (IFCVV) parse_t30_response, &t30p},
	{"+FTSI", (IFCVV) parse_id_response, tsi, NULL},
	{"+FCSI", (IFCVV) parse_id_response, tsi, NULL},
	{"+FCIG", (IFCVV) parse_id_response, tsi, &ptsi},
	{"+FPOLL", (IFCVV) parse_flag,  &poll},
	{NULL}
    };


static PSW psw2[] = {
	{"+FHS", (IFCVV) parse_int_response, &hsc},
	{"+FPS", (IFCVV) parse_int_response, &ppr, &fline},
	{"+FCS", (IFCVV) parse_t30_response, &t30p},
	{"+FIS", (IFCVV) parse_t30_response, &x30p},
	{"+FTC", (IFCVV) parse_t30_response, &t30p},
	{"+FTI", (IFCVV) parse_id_response, tsi, NULL},
	{"+FCI", (IFCVV) parse_id_response, tsi, NULL},
	{"+FPI", (IFCVV) parse_id_response, tsi, &ptsi},
	{"+FPO", (IFCVV) parse_flag,  &poll},
	{NULL}
    };

static PSW *parse_switch[] = {psw0,psw2};

int ParseResponse(char *reply)
{
    struct pswitch *p;

    for(p = parse_switch[c20]; p->string; p++)
    {
	if (strncmp(p->string, reply, strlen(p->string)) == 0)
        {
	    if (p->func != NULL)
            {
		if ((*p->func)(reply, p->ptr, p->extra) < 0)
                {
                    msg(0, " ** Error (%s) ", p->string);
	        }
	    }
	    return (0);
	}
    }
   return (-1);
}


#ifdef TEST
#include <stdarg.h>

int hsc,ppr;
char tsi[64];
short c20 = 0;
T30PARAMS t30p;
int good, poll, ptsi, fline;
char tsi[];
T30PARAMS x30p;

short msg (short code, char *s,...)
{

    char *t, txt[132];
    va_list ap;
    va_start (ap, s);
    t = txt;
    vsprintf (t, s, ap);
    va_end (ap);
    puts (txt);
    return code;
}

logs(char *p)
{
    puts(p);
}
int main(void)
{
    char buf[80] = "+FTSI:\"          0489 12344567 \"\r\n";
    ParseResponse("+FHNG:40\r\n");
    ParseResponse("+FPTS:1\r\n");
    ParseResponse(buf);
    printf("Res %d %d <%s>\n", hsc, ppr, tsi);
    c20 = 1;
    ParseResponse("+FHS:50\r\n");
    printf("Res %d %d <%s>\n", hsc, ppr, tsi);
    return 0;
}
#endif
