#include <exec/types.h>
#include <exec/exec.h>
#include <intuition/intuition.h>
#include <devices/serial.h>
#include <graphics/gfxbase.h>
#include <hardware/cia.h>
#include <dos/dostags.h>
#include <dos/dos.h>

#include <clib/exec_protos.h>
#include <clib/dos_protos.h>
#include <clib/intuition_protos.h>
#include <clib/alib_protos.h>

extern struct GfxBase *GfxBase;

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>

#include "text2g3.h"

#define WFMT "RAW:40/40/%ld/%ld/text2g3 " FAXVERS "/NOSIZE/INACTIVE/PLAIN/SCREEN*"

static struct fmsg
{
    char *prompt;
    short xp;
    short yp;
    short clr;
} f[] = {   {"File  : ", 0, 0, 1},
            {"Format: ", 0, 1, 0}, {"Size : ", 36, 1, 1},
            {"Page  : ", 0, 2, 0}, {"Line : ", 36, 2, 0},
            {"Extra : ", 0, 3, 1},
            {"Msg   : ", 0, 4, 1}
        };
 
#define NMSG (sizeof(f)/sizeof(struct fmsg))

static void PosCsr(short x, short y)
{
    printf("\x9b%d;%dH", y+1, x+1);
}

void InitWindow (short dummy)
{
    struct Screen *scn;
    struct TextFont *tf = NULL;
    char WnName[128];
    short xs, ys;
    short i;
    
    if (scn = (struct Screen *) LockPubScreen (NULL))
    {
	if (tf = (struct TextFont *) GfxBase->DefaultFont)
	{
	    xs = 60 * tf->tf_XSize + scn->WBorLeft + scn->WBorRight;
	    ys = 6 * tf->tf_YSize + 1 + scn->Font->ta_YSize + scn->WBorTop + scn->WBorBottom;
	}
	UnlockPubScreen (NULL, scn);
    }

    sprintf (WnName, WFMT, xs, ys);
    freopen(WnName, "w+", stdout);
    setvbuf(stdout, NULL, _IONBF, 0);
    fputs ("\x1b[0 p\x9b?7l", stdout);
    fputs ("\x1b[42;1;33;>2m\x1b[2J", stdout);
    for(i = 0; i < NMSG; i++)
    {
        PosCsr(f[i].xp, f[i].yp);
        fputs(f[i].prompt, stdout);
    }
    fputs ("\x1b[22;39m", stdout);
}
void Update (short act, char *fmt,...)
{
    va_list ap;
    va_start (ap, fmt);

    PosCsr(f[act].xp + strlen(f[act].prompt), f[act].yp);
    vprintf (fmt, ap);
    if(f[act].clr)
    {
        fputs("\x1b[0K", stdout);
    }
    va_end (ap);
}
void MyExit(short wait, int n)
{
    BPTR fh;

    if(wait)
    {
        fh = (BPTR) fdtofh(fileno(stdout));
        SetMode (fh, 1);
        if (WaitForChar (fh, 5000000))
        {
            char c;
            Read(fh, &c, 1);
        }
        SetMode (fh, 0);
    }
    exit(n);        
}
