/*
 * Manglefont_c, mangles QFX_FONTs.
 * Use as
 *   manglefont act infile outfile
 *
 * where act may include
 *      i - make italic
 *      b - make bold
 *      u - underline
 *
 *  infile  is the input font file
 *  outfile is the output font file
 *
 *  e.g.
 *
 *  ex manglefont;'ib qfax_font qfax-bold-oblique_font'
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#define FONT_YSIZ   28
#define FONT_SIZE   (256*FONT_YSIZ*2)

#define ITALIC  1
#define BOLD    2
#define ULINE   4

main (int ac, char **av)
{
    int fd1, fd2;
    short act = 0;

    if (ac != 4)
    {
	puts ("manglefont act infile outfile");
    }

    if (strchr (av[1], 'i'))
    {
	act |= ITALIC;
    }
    if (strchr (av[1], 'b'))
    {
	act |= BOLD;
    }
    if (strchr (av[1], 'u'))
    {
	act |= ULINE;
    }

    fd1 = open (av[2], O_RDONLY);
    fd2 = open (av[3], O_CREAT | O_WRONLY);

    if (fd1 > 0 && fd2 > 0)
    {
	unsigned char *p, *q;
	short i, j;

	q = (unsigned char *) malloc (FONT_SIZE);
	if (q)
	{
            p = q;
	    read (fd1, q, FONT_SIZE);
	    for (i = 0; i < 256; i++, p += 56)
	    {
		if (act & (BOLD | ITALIC))
		{
                    unsigned char *s;
		    for (s = p, j = 0; j < FONT_YSIZ; j++, s += 2)
		    {
			short r;

                        r = *((short *) s);
			if ((act & ITALIC))
			{
			    if (j < 5)
			    {
				*((short *) s) = (r >> 2);
			    }
			    else if (j < 10)
			    {
				*((short *) s) = (r >> 1);
			    }
			    else if (j > 17 && j < 23)
			    {
				*((short *) s) = (r << 1);
			    }
			    else if (j > 22)
			    {
				*((short *) s) = (r << 2);
			    }
			}

			if ((act & BOLD))
			{
                            r = *((short *)s);
			    *((short *) s) |= (r >> 2);
			}
		    }
		}
		if (act & ULINE)
		{
		    *((unsigned long *) (p + 50)) = 0xffffffff;
		}
	    }
	    write (fd2, q, FONT_SIZE);
	    free (q);
	}
	else
	{
	    perror ("malloc");
	}
	close (fd1);
	close (fd2);
    }
    else
    {
	printf ("%d %d ", fd1, fd2);
	perror ("open");
    }
}
