
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#ifdef _DCC
# include <alloca.h>
#else
# include <unistd.h>
# ifndef __VMS
#  include <memory.h>
# endif
#endif

#include "qfaxparse.h"
#include "qheader.h"

#include "text2g3.h"

#ifdef _DCC
# include <clib/exec_protos.h>
# include <getopt.h>
extern int LoadIFF (char *, short, PPICDEF);
# define getpid() FindTask(NULL)
# define DEFAULT_FILE "qfax.txt"
#else
#ifdef QDOS
# include <qdos.h>
# include <qptr.h>
# define DEFAULT_FILE "qfax_txt"
#else
# define DEFAULT_FILE "qfax.txt"
#endif
#endif

#define TRUE	1
#define FALSE	0
#define MAX_FILE 80

#define DEF_FNTX    16
#define DEF_FNTY    28
#define DEF_FONTSIZE 14336	/* (FNTX/8)*FNTY*256 in  */
#define DEF_LINEHEIGHT	 32
#define DEF_LINESPACE_T  4
#define DEF_LINESPACE_B  2
#define DEF_CHARSPACE_R  1

short faxlinepix[3] = {1728, 2048, 2432};
int faxlinepage[2][3] =
{
    {1120, 1370, 0xffffff},
    {2240, 2740, 0xffffff}
};

#define TOP_MARGIN      31

#ifdef QDOS
# define INFOFILE	"qfax_dat"
# define DEFAULTFONT	"qfax_font"
# define DSEP '_'
#else
# define DSEP '/'
#endif
#ifdef _DCC
# define INFOFILE	"s:qfax.dat"
# define DEFAULTFONT	"s:qfax.font"
#endif
#ifdef __VMS
# define INFOFILE	"sys$login:qfax.dat"
# define DEFAULTFONT	"sys$login:qfax.font"
#endif
#ifdef __linux__
# define INFOFILE	".qfax.dat"
# define DEFAULTFONT	".qfax.font"
#endif

#define CLEAN(p1) (p1 + 2 + strspn ((p1 + 2), " ,\t"))

#ifdef SHMOD
extern short shmod(short a, short b);
#define SMOD(a,b) shmod(a,b)
#else
#define SMOD(a,b) (a % b)
#endif

PICDEF PicDef;
CODERSTATE *code;
static short TopAllow = 0;

static int F_page;
static short F_pix;
static short F_char;

extern short tflag;

typedef struct
{
    char flag[8];
    UCHAR xpix;
    UCHAR ypix;
    UCHAR first;
    UCHAR last;
    UCHAR top_l;
    UCHAR bot_l;
    UCHAR right;
    UCHAR mode;
    long  fsize;
    UCHAR font[1];
} QFONT;

static QFONT *qfont = NULL;
static UCHAR *faxline = NULL;
static UCHAR *prop = NULL;

/**************
static UCHAR font[FONTSIZE];
static UCHAR faxline[FNTY + DEF_LINESPACE_T + DEF_LINESPACE_B][MAXFAXLINELEN];
 **************/

static short leftmargin;
static short lineheight;
static int filestack_ptr;

static FILE *filestack[8];
static char g3file[MAX_FILE];

#ifdef VMS
# define DEFDIR "sys$login:"
#endif

#ifdef QDOS
#define DEFDIR "ram2_"
#endif
#ifdef _DCC
#define DEFDIR "t:"
#endif
#ifdef __unix__
#define DEFDIR "."
#endif

static short uchan = 0;
static char alias[40] = {0};
static char destdir[64] = DEFDIR;
static FAXDATA qd;
static short ufont = 0;
static short ulined = 0;
static short bold = 0;
static short oblique = 0;
static char *usefile = NULL;
static short wait = 1;
static short Tpage;

#define MAXFNT 10
#define MAXMAC 16

typedef struct
{
    void *next;
    short nfont;
    char *fontnam;
} FONTLIST;

FONTLIST *fontlist = NULL;

char *macros[MAXMAC] = {NULL};

extern UCHAR *MagBits (UCHAR *, int, UCHAR, int *);

static MHEAD HFile = {0};
static MHEAD HOff = {0};

#ifdef __VMS
#include <fcntl.h>
#include <unixio.h>
#include <unistd.h>
# include <getopt.h>
char *strdup (const char *);

#define bzero(a,b) memset(a,0,b)
int stricmp (const char *s, const char *d);

#define alloca malloc
#define unlink delete
#endif

#ifndef __linux__
# define lopen fopen
#endif
#if defined (__linux__) || defined (__VMS)
void strins (char *d, const char *s)
{
    int len = strlen (s);	/*  # bytes to insert   */
    char *ptr;

    ptr = d + strlen (d);
    while (ptr >= d)
    {
	ptr[len] = ptr[0];
	--ptr;
    }

    while (*s)
	*++ptr = *s++;
}
#endif

#if 0 /*** defined (QDOS) && !defined(LIBC_IS_FIXED) ***/
# define IOSIZE 16*1024
int xfread(void *buf, int n, int m, FILE *fp)
{
    int nx, nr = n*m;
    int nc = 0, nb = 0;

    while(nr)
    {
	if(nr > IOSIZE) 
	    nx = IOSIZE;
	else
	    nx = nr;

	nb = fread(buf, 1, nx, fp);
	if(nb)
	{
	    nr -= nb;
	    buf = (char *)buf + nb;
	    nc += nb;
	}
	else break;
    }
    return nc;
}
int xread(int fd, void *buf, int n)
{
    int nx,nr = n;
    int nc = 0, nb = 0;

    while(nr)
    {
	if(nr > IOSIZE) 
	    nx = IOSIZE;
	else
	    nx = nr;

	nb = read(fd, buf , nx);
	if(nb)
	{
	    nr -= nb;
	    buf = (char *)buf + nb;
	    nc += nb;
	}
	else break;
    }
    return nc;
}
# ifdef fread
#  undef fread
# endif
# ifdef read
#  undef read
# endif
# define fread xfread
# define read xread
#endif

FONTLIST *findfont(short no)
{
    FONTLIST *p;
    for(p = fontlist; p; p = p->next)
    {
	if(no == p->nfont)
	{
	    break;
	}
    }
    return p;
}

void MemErr (void)
{
    Update (U_MISC, "Out of memory !");
    MyExit (wait, 0);
}

static short maxlh = 0;

int readfont (char *fondat)
{
    FILE *fp;
    
    if ((fp = lopen (fondat, "r")) == NULL && *qd.fontdir)
    {
	char *p,pfont[128];
	p = stpcpy(pfont, qd.fontdir);
	if(*(p-1) == DSEP && *fondat == DSEP) 
	{
	    p--;
	}
	else if(*(p-1) != DSEP && *fondat != DSEP)
	{
	    *p++ = DSEP;
	}
	stpcpy(p, fondat);

	fp = lopen(pfont, "r");
    }
    
    if(fp)
    {
	QFONT q;
	size_t mem;

	fread (&q, offsetof(QFONT,font), 1, fp);
	fseek (fp, 0, 0);
	if (memcmp (q.flag, "QfaxFont", 8) == 0)
	{
	    q.fsize = LG (q.fsize);
	    mem = offsetof(QFONT,font) +  q.fsize;
	    if(q.mode == 'P')
	    {
		mem += (q.last - q.first + 1);
	    }

	    qfont = realloc (qfont, mem);

	    if(qfont)
	    {
		fread (qfont, mem, 1, fp);
		qfont->fsize = LG (qfont->fsize);
		if(q.mode == 'P')
		{
		    prop = qfont->font+qfont->fsize;
		}
		else
		{
		    prop = NULL;
		}
	    }
	    else
	    {
		MemErr ();
	    }
	}
	else
	{
	    mem = offsetof(QFONT,font) + DEF_FONTSIZE;
	    if ((qfont = realloc (qfont, mem)))
	    {
		fread (qfont->font, DEF_FONTSIZE, 1, fp);
		qfont->xpix = DEF_FNTX;
		qfont->ypix = DEF_FNTY;
		qfont->top_l = DEF_LINESPACE_T;
		qfont->bot_l = DEF_LINESPACE_B;
		qfont->right = DEF_CHARSPACE_R;
		qfont->first = 0;
		qfont->last = 255;
		qfont->fsize = DEF_FONTSIZE;
		qfont->mode = 0;
		prop = NULL;
	    }
	    else
	    {
		MemErr ();
	    }
	}

	lineheight = qfont->ypix + qfont->top_l + qfont->bot_l;

	if (lineheight > maxlh)
	{
	    faxline = realloc (faxline, lineheight * MAXFAXLINELEN);
	    memset(faxline+maxlh * MAXFAXLINELEN, '\0', 
		   (lineheight-maxlh) * MAXFAXLINELEN); 
	    maxlh = lineheight;
	}
	
	if(faxline == NULL)
	{
	    MemErr ();
	}
	fclose (fp);
	return 1;
    }

    Update (U_MISC, "Font \"%s\" not found", fondat);
    MyExit (wait, 0);
    return 0;
}

void ChangeFont (short nfnt)
{
    FONTLIST *fn;
    
    bold = ulined = oblique = 0;
    if (ufont != nfnt && (fn = findfont(nfnt)))
    {
	readfont (fn->fontnam);
	ufont = nfnt;
    }
}

void ReadMacroFile (char *mf)
{
    char txt[256];
    FILE *fp;

    if ((fp = lopen (mf, "r")) != NULL)
    {
	while (fgets (txt, sizeof (txt) - 1, fp))
	{
	    char *p, *q;
	    int m;

	    if ((p = strchr (txt, '\n')) != NULL)
	    {
		*p = 0;
	    }
	    m = strtol (txt, &q, 16);
	    if (txt != q && *q++ == ':')
	    {
		if (*(macros + m))
		{
		    Update (U_MISC, "!Macro %d redefined", m);
		}
		*(macros + m) = strdup (q);
	    }
	    else
	    {
		Update (U_MISC, "Invalid Macro: %s", txt);
	    }
	}
	fclose (fp);
    }
}

char *strext (char *src, short offset)
{
    char *p, *p1, *p2 = NULL, *p3;
    short i;

    p = alloca (strlen (src) + 1);
    strcpy (p, src);
    for (p1 = p, i = 0; i <= offset && (p2 = strtok (p1, ",")) != NULL;
	 i++, p1 = NULL)
	;
    if (p2)
    {
	p3 = strdup (p2);
    }
    else
    {
	p3 = NULL;
    }
    return p3;
}

char *MyDate (short flag)
{
    time_t t;
    struct tm *tm;
    static char datebuf[32];
    char *p, *q;

    t = time (NULL);
    tm = localtime (&t);

    q = datebuf;
    if (flag)
    {
	if ((p = strext (qd.dstr, tm->tm_wday)))
	{
	    strcpy (datebuf, p);
	    strcat (datebuf, ", ");
	    q += strlen (q);
	    free (p);
	}
    }
    if ((p = strext (qd.mstr, tm->tm_mon)))
    {
	sprintf (q, "%2d %s %d", tm->tm_mday, p, tm->tm_year + 1900);
	free (p);
    }
    return datebuf;
}

int FaxText (char *str)
{
    int c,r;
    uword i, cx, j;
    int roff, bc, bit_off, byt_idx;
    UCHAR *fptr, *bptr;
    char *ip = str;
    short ml = lineheight;
    short oboff;
    
    memset (faxline, 0, maxlh * MAXFAXLINELEN);
    oboff = qfont->ypix / 5;
    
    cx = leftmargin;

    while (*str)
    {
	if (*str == '~')
	{
	    str++;
	    if (*str >= '0' && *str <= '9')
	    {
		int nf;
		nf = strtol(str, &str, 0);
		if(str && *str == '!') str++;
		
		ChangeFont (nf);
		oboff = qfont->ypix / 5;
		if(lineheight > ml)
		{
		    ml = lineheight;
		}
		continue;
	    }
	    else if (*str == 'u')
	    {
		ulined = 0;
		str++;
		continue;
	    }
	    else if (*str == 'U')
	    {
		ulined = 1;
		str++;
		continue;
	    }
	    else if (*str == 'e')
	    {
		bold = 0;
		str++;
		continue;
	    }
	    else if (*str == 'E')
	    {
		bold = 1;
		str++;
		continue;
	    }
	    else if (*str == 'o')
	    {
		oblique = 0;
		str++;
		continue;
	    }
	    else if (*str == 'O')
	    {
		oblique = 1;
		str++;
		continue;
	    }
	    else if (*str == 'D' || *str == 'd')
	    {
		char *p;

		if ((p = MyDate ((*str == 'D'))))
		{
		    strins (++str, p);
		}
	    }
	    else if (*str == 'm')
	    {
		str++;
		if (isxdigit (*str))
		{
		    short n;

		    n = (*str > '9') ? tolower (*str) - 'a' + 0xa : *str - '0';
		    strins (++str, *(macros + n));
		}
	    }
	    else
	    {
		if (*str == '\0')
		{
		    continue;
		}
	    }
	}

	c = *str++;
	c &= 0xFF;

	if(c < qfont->first || c > qfont->last)
	{
	    Update (U_MISC, "Out of range: %d (%d)", c, ufont);
	    c = '*';
	}	    

	if(prop)
	{
	    short ll;
	    UCHAR *pr = prop;
	    short lc = c - qfont->first;
	    
	    for(roff =0, ll = 0; ll < lc; ll++)
	    {
		roff += *pr++;
	    }
	    qfont->xpix = *pr;
	}
	else
	{
	    roff = c * qfont->xpix;     
	}
	

	if (ulined)
	{
	    ulined = ((qfont->ypix - (qfont->ypix >> 3)) + 1) & ~2;
	}

	r = roff * qfont->ypix;

	bptr = faxline + (qfont->top_l) * MAXFAXLINELEN;
	for (i = 0; i < qfont->ypix; )
	{
	    int fbyt, foff;
	    short ps;

	    fptr = bptr;
	    
	    byt_idx = cx >> 3;
	    bit_off = cx & 0x07;

	    fbyt = r >> 3;
	    foff = r & 7;
	    ps = 0;

	    if (ulined && i == ulined)
	    {
		short bb, by;

		bb = byt_idx;
		by = bit_off;
		for (bc = 0; bc < (qfont->xpix+qfont->right); bc++, by++)
		{
		    if (by > 7)
		    {
			by = 0;
			bb++;
		    }
		    fptr[bb] |= code->mbits[by];
		}
	    }

	    if (oblique)
	    {
		short mps = (qfont->xpix >> 3);
		
		if (i < oboff)
		{
		    ps = mps;
		}
		else if (i < (oboff << 1))
		{
		    ps = mps >> 1;
		}
		else if (i >= qfont->ypix - oboff)
		{
		    ps = -mps;
		}
		else if (i >= qfont->ypix - (oboff << 1))
		{
		    ps = -(mps >> 1);
		}
	    }

	    if (ps)
	    {
		bit_off += ps;
		if (bit_off > 7)
		{
		    byt_idx++;
		}
		else if (bit_off < 0)
		{
		    byt_idx--;
		}
		bit_off &= 7;
	    }

	    for (bc = 0; bc < qfont->xpix; bc++, bit_off++, foff++)
	    {
		static UCHAR fmap[] = {0x80,0x40,0x20,0x10,0x8,0x4,0x2,0x1};

		if (bit_off > 7)
		{
		    bit_off &= 7;
		    byt_idx++;
		}
		if (foff > 7)
		{
		    fbyt++;
		    foff &= 7;
		}
		
		if (fmap[foff] & qfont->font[fbyt])
		{
		    fptr[byt_idx] |= code->mbits[bit_off];

		    if (bold)
		    {
			short bb, by;

			bb = byt_idx;
			by = bit_off;
			by += 2;
			if (by > 7)
			{
			    bb++;
			    by &= 7;
			}
			fptr[bb] |= code->mbits[by];
		    }
		}
	    }

	    if (!code->res)
	    {
		i++, bptr += MAXFAXLINELEN, r += qfont->xpix;
	    }
	    i++, bptr += MAXFAXLINELEN, r += qfont->xpix;
	}

	cx += (qfont->xpix + qfont->right);
	if (cx + qfont->xpix > F_pix)
	{
	    if (str && *str)
	    {
		*str = 0;
		if (str > (ip + 40))
		{
		    *(ip + 40) = 0;
		}
		Update (U_MISC, "Truncate: %s", ip);
	    }
	}
    }

    fptr = faxline;
    for (i = 0; i < ml;)
    {
	if (PicDef.active)
	{
	    if ((*PicDef.pixfunc) (&PicDef, code->res))
	    {
		for (j = PicDef.from; j < F_char; j++)
		{
		    fptr[j] |= PicDef.pixlin[j];
		}
	    }
	    else
	    {
		free (PicDef.base);
		PicDef.active = FALSE;
	    }
	}
	FaxCodeLine (fptr);
	if (!code->res)
	{
	    i++, fptr += MAXFAXLINELEN;
	}
	i++, fptr += MAXFAXLINELEN;
    }
    return 0;
}

static short fmt1[3] = {36, 45, 58};
static short fmt2[3] = {20, 30, 40};

void HeadPage (int page, int maxpage)
{
    struct tm *pt;
    char locid[24];
    char text[312];
    time_t t = 0;
    short lm;
    short tm;
    char *p, *q;

    tm = (code->res) ? TOP_MARGIN : TOP_MARGIN >> 1;

    FaxBeginPage ();

    p = qd.LocalId;
    q = locid;

    do
    {
	if (*p != '"')
	{
	    *q++ = *p;
	}
    }
    while (*p++);

    bzero ((char *) text, F_char);
    for (lm = 0; lm < tm; lm++)
    {
	FaxCodeLine ((UCHAR *) text);
    }

    t = time (NULL);
    pt = localtime (&t);
    if(tflag) memset(pt, 0, sizeof(struct tm));
    
    sprintf (text, "From: %-*.*s%-*.*s %02d/%02d/%02d %02d:%02d    P. %03d/%03d",
	     fmt1[(short)qd.width_g3], fmt1[(short)qd.width_g3],
	     qd.Signature,
	     fmt2[(short)qd.width_g3], fmt2[(short)qd.width_g3],
	     locid,
	     pt->tm_mday, pt->tm_mon + 1, (pt->tm_year % 100),
	     pt->tm_hour, pt->tm_min, page, maxpage);

    lm = leftmargin;

    leftmargin = 17 * 4;
    FaxText (text);
    leftmargin = lm;

    bzero ((char *) text, F_char);
    for (lm = 0; lm < tm - 1; lm++)
    {
	FaxCodeLine ((UCHAR *) text);
    }
    FaxCodeAlign ((UCHAR *) text);
    FaxFlushBuf (((code->bitnr + 7) >> 3));
}

void MakeOutFile (int maxpage)
{
    char ifn[64];
    int ifd, nc, i;
    long offset = 0;
    int rt = 1;
    int res;
    
    if (uchan)
    {
	res = FaxOpenFile (NULL, code->res, code->pack, F_pix);
	code->hd = fileno (stdout);
    }
    else
    {
	res = FaxOpenFile (g3file, code->res, code->pack, F_pix);
    }
    if(res)
    {
        Update (U_MISC, "File error:%s %s", g3file, strerror(res));
        MyExit (wait, 0);
    }
    
    code->maxlines += TopAllow;
    FaxHeader (maxpage, qd.width_g3, qd.page_g3, qd.LocalId, 0);

    for (i = 0; i < maxpage; i++)
    {
	struct stat s;

	if (HFile.head)
	{
	    char *s;

	    if ((s = CheckInclude (HFile.head, i, &HOff, code, &qd, &rt)))
	    {
		Update (U_MISC, "Include Error :%s", s);
		MyExit (wait, 0);
	    }
	}
	sprintf (ifn, "%s%x.%d", destdir, getpid (), i + 1);
	ifd = open (ifn, O_RDONLY, 0666);
	if (fstat (ifd, &s) == 0 && s.st_size)
	{
	    if (code->hd < 0 || ifd < 0)
	    {
		Update (U_MISC, "File error:%d %d %d",code->hd, ifd, errno);
		MyExit (wait, 0);
	    }

	    if (HOff.tail)
	    {
		OFFREC *o;

		o = (OFFREC *) (HOff.tail)->udata;
		if (o->code != code->pack ||
		    o->res != code->res || o->page != qd.page_g3)
		{
		    MLIST *ml;

		    ml = malloc (sizeof (MLIST *) + sizeof (OFFREC));
		    o = (OFFREC *) ml->udata;
		    o->pageno = rt;
		    o->off = lseek (code->hd, 0, 1);
		    o->code = code->pack;
		    o->res = code->res;
		    o->page = qd.page_g3;
		    o->width = qd.width_g3;

		    AppendList (&HOff, ml);
		}
	    }
	    HeadPage (i + 1, maxpage);
	    while ((nc = read (ifd, code->outbuf, OUTBUFSIZE)) > 0)
	    {
		if (write (code->hd, code->outbuf, nc) != nc)
		{
		    Update (U_MISC, "System: Write error 3");
		    MyExit (wait, 0);
		}
	    }
	    rt += 1;
	}
	close (ifd);
	unlink (ifn);
    }
    if (HFile.head)
    {
	char *s;

	if ((s = CheckInclude (HFile.head, i, &HOff, code, &qd, &rt)))
	{
	    Update (U_MISC, "Include Error %s", s);
	    MyExit (wait, 0);
	}
    }
    if (HOff.head)
    {
	MLIST *m;

	offset = lseek (code->hd, 0, 1);
	for (m = HOff.head; m; m = m->next)
	{
#if 0 
	    fprintf (stderr, "Offset %d %d %d\n",
		     ((OFFREC *) m->udata)->off,
		     ((OFFREC *) m->udata)->res,
		     ((OFFREC *) m->udata)->code);
#endif
	    OffrecEndianise((OFFREC *) m->udata);
	    write (code->hd, m->udata, sizeof (OFFREC));
	}
    }
    FaxHeader (rt - 1, qd.width_g3, qd.page_g3, qd.LocalId, offset);
#ifdef QDOS
    if (!uchan)
#endif
	close (code->hd);
    code->hd = 0;
}

static char outfn[64];

void FaxOpen (int page)
{
    int res;
    
    sprintf (outfn, "%s%x.%d", destdir, getpid (), page);

    if ((res = FaxOpenFile (outfn, code->res, code->pack, F_pix)) == 0)
    {
	Update (U_PAGE, "%3d", page);
    }
    else
    {
	Update (U_MISC, "Can't open output file %s", outfn, strerror(res));
	MyExit (wait, -3);
    }
}


void build_dst_name (char *filename)
{
    char *pp;
    char *ps = NULL;
    int use_fax = FALSE;

    if (usefile)
    {
	ps = usefile;
	use_fax = TRUE;
    }
    else
    {
#ifdef QDOS
	if (isdirdev (filename))
	{
	    ps = strchr (filename, '_');
	}
#else
	ps = strrchr (filename, '/');
	if (ps == NULL)
	{
	    ps = strrchr (filename, ':');
	}
#endif
	if (ps == NULL)
	{
	    ps = filename;
	}
	else
	{
	    ps++;
	}
    }
    pp = stpcpy (g3file, destdir);
    strcpy (pp, ps);
    if (usefile == NULL)
    {
#ifdef QDOS
	if ((pp = strrchr (g3file, '_')) != NULL)
#else
	if ((pp = strrchr (g3file, '.')) != NULL)
#endif
	{
	    *pp = 0;
	}
    }
    strcat (g3file, qd.suffix);
}

void MagIT (PPICDEF p)
{
    int nb = 0x80;
    UCHAR *t, *d;
    short i;

    d = (UCHAR *) p->pixlin + p->from;
    bzero ((char *) p->pixlin, F_char);
    t = (UCHAR *) p->ptr;
    for (i = 0; i < p->llen; i++, t++)
    {
	d = MagBits (d, p->mag, *t, &nb);
    }
    p->ptr += p->llen;
    p->line++;
}

int GetBitMap (PPICDEF p, short xxx)
{
    UCHAR *d;

    if (p->line >= p->ypix)
    {
	return 0;
    }

    if (p->mag == 1)
    {
	d = (UCHAR *) p->pixlin + p->from;
	bzero ((char *) p->pixlin, F_char);
	memcpy (d, p->ptr, p->llen);
	p->ptr += p->llen;
	p->line++;
	if (code->res == 0)
	{
	    p->ptr += p->llen;
	    p->line++;
	}
    }
    else
    {
	if (code->res || p->mag == 2)
	{
	    if (p->magc == 0)
	    {
		MagIT (p);
	    }
	    p->magc++;
	    if (code->res == 0 && p->mag == 2)
	    {
		p->magc++;
	    }
	    p->magc = SMOD (p->magc,p->mag);
	}
	else
	{
	    if (SMOD(p->magc,3) != 1)
	    {
		MagIT (p);
	    }
	    p->magc++;
	}
    }
    return 1;
}

short DoNewPage (short page)
{
    FaxCodeEOP ();
    FaxCloseFile ();
    page++;
    FaxOpen (page);
    return page;
}

void IncludeGraphic (PPICDEF p, short overlay, short *page, short *line)
{
    short pg, n = 0;

    if (code->scan + p->faxlines > F_page)
    {
	pg = *page;
	Update (U_MISC, "Graphic forces new page");
	pg = DoNewPage (pg);
	*page = pg;
	*line = 0;
    }

    if (overlay)
    {
	p->active = 1;
    }
    else
    {
	p->active = 0;
	while ((*p->pixfunc) (p, code->res))
	{
	    n++;
	    FaxCodeLine ((UCHAR *) p->pixlin);
	}
	free ((void *) p->base);
    }
}

int LoadPBM (char *cmd, PPICDEF p)
{
    FILE *fp;
    int numparm;
    char sfn[256];
    short from;
    char *sp;
    short mag,centre = 0, i, x, sw;

    if (p->active)
    {
	Update (U_MISC, "Can't overlay graphics");
	return -1;
    }

    while ((sp = strchr (cmd, ',')))
    {
	*sp = ' ';
    }

    numparm = sscanf (cmd, "%s%hd%hd", sfn, &from, &mag);

    switch (numparm)
    {
	case 0:
	    return -1;
	    break;
	case 1:
	    from = 160;
	    mag = 1;
	    break;
	case 2:
	    mag = 1;
	    break;
	default:
	    if (mag < 0)
	    {
		mag = -mag;
		centre = TRUE;
	    }
	    if (mag > 3)
	    {
		mag = 3;
	    }
	    if (mag < 1)
	    {
		mag = 1;
	    }
	    break;
    }

    p->ptr = NULL;
    p->pixfunc = GetBitMap;
    p->active = FALSE;
    p->line = 0;
    p->magc = 0;

    Update (U_INC, sfn);
    fp = lopen (sfn, "r");
    if (fp == NULL)
    {
	Update (U_MISC, "Can't open PBM file");
	return -1;
    }

    for (i = 0; i < 2;)
    {
	fgets (sfn, sizeof (sfn), fp);
	if (*sfn == '#')
	{
	    char *ap;

	    if ((ap = strchr (sfn, '\n')))
	    {
		*ap = 0;
	    }
	    Update (U_MISC, sfn);
	}
	else
	{
	    if (i == 0)
	    {
		if (strncmp (sfn, "P4", 2))
		{
		    fclose (fp);
		    Update (U_MISC, "Not a PBM file");
		    return -1;
		}
		else
		{
		    i++;
		}
	    }
	    else if (i == 1)
	    {
		x = strtol (sfn, &sp, 10);
		p->ypix = strtol (sp, NULL, 10);
		p->llen = (x + 7) >> 3;
		i = 2;
	    }
	}
    }

    sw = mag * p->llen;

    from = (from + 7) >> 3;
    if (centre)
    {
	from = (F_char - sw) >> 1;
    }
    else if (from + sw > F_char)
    {
	from = F_char - sw;
    }
    if (from < 0)
    {
	from = 4;
    }

    if (from + sw > F_char)
    {
	Update (U_MISC, "Picture is too big");
	fclose (fp);
	return -1;
    }

    p->from = from;
    p->ptr = p->base = malloc (p->llen * p->ypix);
    p->mag = mag;

    if (p->base)
    {
	short j;

	sp = p->base;
	for (j = 0; j < p->ypix; j++)
	{
	    fread (sp, p->llen, 1, fp);
	    sp += p->llen;
	}
	p->faxlines = p->ypix >> (1 - code->res);
    }
    fclose (fp);
    return 0;
}

int LoadQLPix (char *cmd, int sflag)
{

    int i;
    short sf, mag, from, numparm;
    char *sp, sfn[MAX_FILE];
    char centre = 0;
    short sw, square = 0, invert = 0;

    if (PicDef.active)
    {
	Update (U_MISC, "Can't overlay graphics");
	return -1;
    }


    while ((sp = strchr (cmd, ',')))
    {
	*sp = ' ';
    }

    numparm = sscanf (cmd, "%s%hd%hd%hd%hd", sfn, &from, &mag, &invert, &square);

    
    if (numparm < 4)
    {
	invert = 0;
    }

    if (numparm < 5)
    {
	square = 0;
    }

    switch (numparm)
    {
	case 0:
	    return -1;
	    break;
	case 1:
	    from = 160;
	    mag = 1;
	    break;
	case 2:
	    mag = 1;
	default:
	    if (mag < 0)
	    {
		mag = -mag;
		centre = TRUE;
	    }
	    if (mag > 3)
	    {
		mag = 3;
	    }
	    if (mag < 1)
	    {
		mag = 1;
	    }
	    break;
    }

    if ((sf = open (sfn, O_RDONLY, 0666)) > 0)
    {

	struct _qhead_
	{
	    unsigned short flag;
	    unsigned short xpix;
	    unsigned short ypix;
	    unsigned short llen;
	    unsigned short zero;
	}
	qhead;
	char *sbuff;
	struct stat s;

	Update (U_INC, sfn);

	if (sflag)
	{
	    read (sf, &qhead, sizeof (qhead));
#if BYTE_ORDER == LITTLE_ENDIAN
#define SWAPW(p1) p1=SH(p1)
	    SWAPW (qhead.flag);	    
	    SWAPW (qhead.xpix);	    
	    SWAPW (qhead.ypix);	    
	    SWAPW (qhead.llen);	    
	    SWAPW (qhead.zero);	    
#endif
	    if (qhead.flag != 0x4afc)
	    {
		Update (U_MISC, "Not a PIC file %x", qhead.flag);
		close (sf);
		return -1;
	    }
	    sbuff = malloc ((i = qhead.ypix * qhead.llen));
	}
	else
	{
	    fstat (sf, &s);
	    qhead.xpix = 512;
	    qhead.llen = 128;
	    qhead.ypix = (s.st_size + 127) >> 7;
	    sbuff = malloc ((i = s.st_size));
	}

	if (sbuff == NULL)
	{
	    Update (U_MISC, "No memory");
	    close (sf);
	    return -1;
	}

	read (sf, sbuff, i);

	sw = ((qhead.xpix + 7) >> 3) * mag;
	from = (from + 7) >> 3;
	if (centre)
	{
	    from = (F_char - sw) >> 1;
	}
	else if (from + sw > F_char)
	{
	    from = F_char - sw;
	}
	if (from < 0)
	{
	    from = 4;
	}

	if (from + sw > F_char)
	{
	    Update (U_MISC, "Picture is too big");
	    free (sbuff);
	    close (sf);
	    return -1;
	}

	PicDef.base = sbuff;
	PicDef.ptr = NULL;
	PicDef.from = from;
	PicDef.ypix = qhead.ypix;
	PicDef.llen = qhead.llen;
	PicDef.square = square;
	PicDef.invert = invert;
	PicDef.xpix = qhead.xpix;
	PicDef.mag = mag;
	PicDef.magc = PicDef.aspect = PicDef.line = 0;
	PicDef.pixfunc = GetQLRec;
	PicDef.faxlines = (PicDef.ypix * mag) >> (1 - code->res);

	if (!square)
	{
	    PicDef.faxlines = (PicDef.faxlines << 2) / 3;
	}
#ifndef QDOS
	Update (U_MISC, "QGraph (%d) %d %d %d",
		numparm,
		PicDef.mag,
		PicDef.square,
		PicDef.faxlines);
#endif
    }
    else
    {
	Update (U_MISC, "could not read %s", sfn);
	
	return -1;
    }
    return 0;
}

void AddFontList (char *s, short num)
{
    FONTLIST *fn;
    
    if (num >= 0)
    {
	if ((fn = findfont(num)))
	{
	    free (fn->fontnam);
	}
	else
	{
	    fn = malloc(sizeof(FONTLIST));
	    fn->nfont = num;
	    fn->next = fontlist;
	    fontlist = fn;
	}
	fn->fontnam = strdup(s);	
    }
}

void GetNewFont (char *cmd)
{
    short fnt;
    char *sp;

    while ((sp = strchr (cmd, ',')))
    {
	*sp = ' ';
    }
    sp = strchr (cmd, ' ');
    if (sp)
    {
	*sp++ = '\0';
	fnt = atoi (sp);
	AddFontList (cmd, fnt);
    }
}

static short charht;

void DoText (char *p, short *page, short *line)
{
    char txt[16];

    if ((code->scan + charht) > F_page)
    {
	*page = DoNewPage (*page);
	*line = 0;
    }

    (*line)++;
    sprintf (txt, "%5d", *line);
    Update (U_LINE, txt);
    FaxText (p);
}

void SetPageSize (void)
{
    char ptext[132];

    F_page = faxlinepage[(short)code->res][(short)qd.page_g3];
    TopAllow = (lineheight + TOP_MARGIN * 2) >> (1 - code->res);
    F_pix = faxlinepix[(short)qd.width_g3];
    F_char = F_pix >> 3;
    sprintf (ptext, "%d x ", faxlinepix[(short)qd.width_g3]);
    if (F_page == 0xFFFFFF)
    {
	strcat (ptext, "unlimited");
    }
    else
    {
	sprintf (ptext + strlen (ptext), "%d", F_page);
    }
    Update (U_STAT, ptext);

    F_page -= TopAllow;
    code->pix = F_pix;
}

void CheckActive(void)
{
    if (PicDef.active)
    {
	while ((*PicDef.pixfunc) (&PicDef, code->res))
	{
	    memset (faxline , '\0', PicDef.from);    
	    memcpy (faxline + PicDef.from,
		    PicDef.pixlin+PicDef.from,
		    F_char - PicDef.from);
	    FaxCodeLine (faxline);
	}
	free (PicDef.base);
	PicDef.active = FALSE;
    }	
}

int convert (char *filename)
{
    FILE *F;

    short line = 0;
    short done = 0;
    char str[256], *sptr;

    Tpage = 1;
    charht = lineheight >> (1 - code->res);
    filestack_ptr = 0;
    ufont = 0;

    build_dst_name (filename);

    F = lopen (filename, "r");
    if (F)
    {
	short cont = 0;
	short first = 0;

	Update (U_IFILE, filename);

	FaxOpen (Tpage);

	while (!done)
	{
	    char *pstr;

	    sptr = fgets (str, sizeof (str), F);
	    if (sptr == NULL)
	    {
		if (filestack_ptr == 0)
		{
		    done = 1;
		    break;
		}
		else
		{
		    filestack_ptr--;
		    fclose (F);
		    F = filestack[filestack_ptr];
		    continue;
		}
	    }

	    if ((pstr = strchr (str, '\n')))
	    {
		*pstr = '\0';
	    }

	    do
	    {
		cont = 0;
		pstr = NULL;
		if (*sptr == '~' && *(sptr + 1) != '~')
		{
		    char act;

		    switch ((act = *(sptr + 1)))
		    {
			case '#':
			    if (first == 0)
			    {
				char *sx;

				qd.width_g3 = strtol (sptr + 2, &sx, 10);
				qd.page_g3 = strtol (sx, NULL, 10);
				if (qd.width_g3 < 0 || qd.width_g3 > 2)
				{
				    qd.width_g3 = 0;
				}
				if (qd.page_g3 < 0 || qd.page_g3 > 2)
				{
				    qd.page_g3 = 0;
				}
				SetPageSize ();
				first = 1;
			    }
			    break;
			case 'm':
			    if (isxdigit (*(sptr + 2)))
			    {
				short n = 15;

				n = (*(sptr + 2) > '9') ? tolower (*(sptr + 2)) - 'a' + 0xa : *(sptr + 2) - '0';
				strins (sptr + 3, *(macros + n));
				if (*(sptr + 3) == '~')
				{
				    sptr += 3;
				    cont = 1;
				}
				else
				{
				    pstr = ++sptr;
				}
			    }
			    break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			    {
				char *q;
				int nf;
				nf = strtol(sptr+1, &q, 0);
				if(q && *q == '!') q++;
				
				ChangeFont (nf);
				sptr = q - 2;
				pstr = sptr;
			    }
			    
			    break;

			case 'D':
			case 'd':
			    first = 1;
			    {
				char *p;

				if ((p = MyDate ((act == 'D'))))
				{
				    strins (sptr + 2, p);
				}
			    }
			    pstr = sptr;
			    break;
			case 'u':
			    ulined = 0;
			    if (*(sptr + 2) == '~')
			    {
				sptr += 2;
				cont = 1;
			    }
			    else
			    {
				pstr = sptr;
			    }
			    break;

			case 'U':
			    ulined = 1;
			    if (*(sptr + 2) == '~')
			    {
				sptr += 2;
				cont = 1;
			    }
			    else
			    {
				pstr = sptr;
			    }
			    break;

			case 'e':
			    bold = 0;
			    if (*(sptr + 2) == '~')
			    {
				sptr += 2;
				cont = 1;
			    }
			    else
			    {
				pstr = sptr;
			    }
			    break;

			case 'E':
			    bold = 1;
			    if (*(sptr + 2) == '~')
			    {
				sptr += 2;
				cont = 1;
			    }
			    else
			    {
				pstr = sptr;
			    }
			    break;

			case 'o':
			    oblique = 0;
			    if (*(sptr + 2) == '~')
			    {
				sptr += 2;
				cont = 1;
			    }
			    else
			    {
				pstr = sptr;
			    }
			    break;

			case 'O':
			    oblique = 1;
			    if (*(sptr + 2) == '~')
			    {
				sptr += 2;
				cont = 1;
			    }
			    else
			    {
				pstr = sptr;
			    }
			    break;

			case 'F':
			    GetNewFont (CLEAN (sptr));
			    break;
			case 'P':
			    CheckActive();
			    first = 1;
			    Tpage = DoNewPage (Tpage);
			    line = 0;
			    break;
			case 'I':	/* Include this File */
			    if (filestack_ptr < 7)	/* limit not exceeded */
			    {
				filestack[filestack_ptr++] = F;
				F = lopen (CLEAN (sptr), "r");
				if (F)
				{
				    Update (U_INC, sptr + 2);
				}
				else
				{
				    Update (U_MISC, "could not include <%s>", sptr + 2);
				    F = filestack[--filestack_ptr];
				}
			    }
			    else
			    {
				Update (U_MISC, "include limit exceeded, command ignored");
			    }
			    break;
			case 'C':	/* chain to this File */
			    {
				FILE *newfile;

				newfile = lopen (CLEAN (sptr), "r");
				if (newfile)
				{
				    Update (U_INC, sptr + 2);
				    fclose (F);
				    F = newfile;
				}
				else
				{
				    Update (U_MISC, "could not chain to %s", sptr + 2);
				}
			    }
			    break;
			case 'S':	/* include raw screenfile */
			case 's':	/* overlay raw screenfile */
			case 'T':	/* overlay raw screenfile */
			    if (LoadQLPix (CLEAN (sptr), 0) == 0)
			    {
				IncludeGraphic (&PicDef, (act != 'S'), &Tpage, &line);
			    }
			    first = 1;
			    break;
#ifdef _DCC
			case 'A':	/* include IFF  */
			case 'a':	/* overlay IFF  */
			    if (LoadIFF (CLEAN (sptr), code->res, &PicDef) == 0)
			    {
				IncludeGraphic (&PicDef, (act == 'a'), &Tpage, &line);
			    }
			    first = 1;
			    break;
#endif
			case 'B':	/* include pbm  */
			case 'b':	/* overlay pbm  */
			    if (LoadPBM (CLEAN (sptr), &PicDef) == 0)
			    {
				IncludeGraphic (&PicDef, (act == 'b'), &Tpage, &line);
			    }
			    first = 1;
			    break;

			case 'Q':	/* include QPaint */
			case 'q':	/* overlay QPaint */
			    if (LoadQLPix (CLEAN (sptr), 1) == 0)
			    {
				IncludeGraphic (&PicDef, (act != 'Q'), &Tpage, &line);
			    }
			    first = 1;
			    break;
			case ';':	/* Comment */
			    break;
			case ':':	/* Verbose comment */
			    {
				char *p, *q;

				p = CLEAN (sptr);
				if ((q = strchr (p, '\n')))
				{
				    *q = 0;
				}
				Update (U_MISC, ": %s", p);
			    }
			    break;
			case 'X':
			case 'x':
			    {
				char *p;
				CheckActive();
				p = CLEAN (sptr);
				if (p && *p)
				{
				    MLIST *m;
				    FLIST *f;

				    if (code->scan)
				    {
					Tpage = DoNewPage (Tpage);
					line = 0;
				    }

				    m = malloc (sizeof (MLIST *) + sizeof (FLIST));
				    f = (FLIST *) m->udata;
				    f->ap = Tpage - 1;
				    strcpy (f->fnam, p);
				    AppendList (&HFile, m);
				}
				else
				{
				    done = 1;
				}
			    }
			    break;
			default:
			    Update (U_MISC, "??: %s", sptr + 1);
			    break;
		    }
		    if (cont == 0 && pstr)
		    {
			if (*(sptr + 2) && *(sptr + 2) != '\n')
			{
			    DoText (sptr + 2, &Tpage, &line);
			    first = 1;
			}
		    }
		}
		else if (sptr)
		{
		    DoText (str, &Tpage, &line);
		    first = 1;
		}
	    }
	    while (cont);
	}
	CheckActive();
	fclose (F);
	if (code->hd)
	{
	    if (code->scan)
	    {
		FaxCodeEOP ();
	    }
	    else
	    {
		Tpage--;
	    }
	    FaxCloseFile ();
	}

	Update (U_MISC, "Adding Headers");
	ChangeFont (0);
	oblique = bold = ulined = 0;

	MakeOutFile (Tpage);
	return (0);
    }
    else
    {
	Update (U_MISC, "Can't open input file %s", filename);
	if (code->hd)
	{
	    close (code->hd);
	    unlink (outfn);
	}
	return (-1);
    }
}

int init_myfax (void)
{
    static char *mm = "January,February,March,April,May,June,July,August,September,October,November,December";
    static char *dd = "Sunday,Monday,Tuesday,Wednesday,Thursday,Friday,Saturday";

    PicDef.active = FALSE;

    FaxParams (&qd);
    qd.elsb[0] = 1;

    if (*qd.Font)
    {
	AddFontList (qd.Font, 0);
    }
    if (*qd.Temp)
    {
	strcpy (destdir, qd.Temp);
    }

    if (qd.dstr == NULL)
    {
	qd.dstr = dd;
    }
    if (qd.mstr == NULL)
    {
	qd.mstr = mm;
    }

    return 1;
}
#ifdef QDOS
int (*_cmdwildcard) (char *, char ***, int *) = cmdexpand;

#endif
#ifdef _DCC
extern int WildCards (int *, char ***);

#endif

void CleanUp (void)
{
    char ifn[64];
    int i;

    if (!uchan && code->hd)
    {
	close (code->hd);
	unlink (g3file);
	code->hd = 0;
    }

    for (i = 0; i <= Tpage; i++)
    {
	sprintf (ifn, "%s%x.%d", destdir, getpid (), i + 1);
	unlink (ifn);
    }

}

static ALIAS *FindAlias (char *nam)
{
    static ALIAS *pal = NULL;

    if (qd.alias)
    {
	ALIAS *pl;

	for (pl = qd.alias; pl && !pal; pl = pl->next)
	{
	    if (strcasecmp (nam, pl->abbrev) == 0 ||
		strcasecmp (nam, pl->alias) == 0)
	    {
		pal = pl;
	    }
	}
    }
    return pal;
}

int main (int ac, char **av)
{
    int c, st, et,i;
    extern char *optarg;
    extern int optind;
    short errflg = 0;
    int res, pack;
    short noptr = 0;

#ifdef _DCC
    WildCards (&ac, &av);
#endif

    for(i = 1; i < ac; i++)
    {
	char *p = NULL;

	if (**(av+i) == '-')
	{
	    if (*(1+*(av+i)) == 'C')
	    {
		if (*(2+*(av+i)))
		{
		    p = (2+*(av+i));
		}
		else if (*(av + 1 + i) && **(av + 1 + i) != '-')
		{
		    p = *(av+i+1);
		}
		if (p)
		{
		    char t[80];
		    if(access(p, R_OK) == 0)
		    {
			sprintf(t,"QFAX_DATA=%s",p);
			putenv(t);
		    }
		}
		break;
	    }
	}
    }

    atexit (CleanUp);
    st = time (NULL);
    if (init_myfax ())
    {
	res = qd.res;
	pack = 0;

	if (qd.t2g3ini)
	{
	    FrigArgs (&ac, &av, qd.t2g3ini);
	}

	while ((c = getopt (ac, av, "a:l:Mn12if:o:xr:m:u:w:h:g:OTC:")) != -1)
	{
	    switch (c)
	    {
		case 'g':
		    qd.tg3 = atoi (optarg);
		    break;
		case 'a':
		    strcpy (alias, optarg);
		    break;
		case 'M':
		    qd.elsb[0] = 0;
		    break;
		case 'm':
		    ReadMacroFile (optarg);
		    break;
		case '2':
		    pack = 1;
		    break;
		case '1':
		    pack = 0;
		    break;
		case 'r':
		    res = (atoi (optarg) != 0);
		    break;
		case 'x':
#ifdef QDOS
		    {
			extern char *_endmsg;

			_endmsg = NULL;
		    }
#endif
		    wait = 0;
		    break;
		case 'o':
		    strcpy (destdir, optarg);
		    break;
		case 'f':
		    AddFontList (optarg, 0);
		    break;
		case 'u':
		    usefile = strdup (optarg);
		    break;
		case 'w':
		    qd.width_g3 = atoi (optarg);
		    break;
		case 'h':
		    qd.page_g3 = atoi (optarg);
		    break;
		case 'l':
		    qd.lmargin = atoi (optarg);
		    break;
		case '?':
		    errflg++;
		    break;
		case 'n':
		    noptr = 1;
		    break;
                case 'T':
                    tflag = 1;
                    break;
	    }
	}

	code = FaxInitCoder (qd.elsb[0]);
	code->res = res;
	code->gflag = qd.tg3;
	leftmargin = qd.lmargin;
#ifndef QDOS
	InitWindow (noptr);
#else
	if (isatty (fileno (stdin)))
	{
	    InitWindow (noptr, ((*qd.mono) & 1));
	}
	uchan = !isatty (fileno (stdout));
#endif
        {
            
            char *pp = destdir + strlen(destdir) - 1;
            if(pp != destdir && *pp != DSEP)
            {
                *(++pp) = DSEP;
                *(++pp) = 0;
            }
        }
        
	if (*alias)
	{
	    short tempcod;
	    ALIAS *pal;

	    if ((pal = FindAlias (alias)))
	    {
		tempcod = pal->flag & 1;
		if (tempcod != pack)
		{
		    Update (U_MISC, "Using coding from Phone Book");
		    pack = tempcod;
		}
	    }
	}

	code->pack = pack;

	if (fontlist == NULL)
	{
	    AddFontList (DEFAULTFONT, 0);
	}

	readfont (fontlist->fontnam);
	if (errflg)
	{
	    Update (U_MISC, "text2g3 [-x] [-r resolution (0/1)] [-f fontfile0] [-o outdir] file ... fileN\n");
	    MyExit (wait, 0);
	}


	Update (U_INFO, "%s res, %s",
		(code->res) ? "fine" : "normal",
		(code->pack) ? "2D-Read" : "1-D Huffman");

	SetPageSize ();

	oblique = bold = ulined = 0;
	if (optind < ac)
	{
	    int ind;

	    for (ind = optind; ind < ac; ind++)
	    {
		if ((res = convert (av[ind])))
		{
		    break;
		}
	    }
	}
	else
	{
	    res = convert (DEFAULT_FILE);
	}

	et = time (NULL);
	if (!res)
	{
	    Update (U_MISC, "That took %d seconds; ", et - st);
	}
    }
    MyExit (wait, 0);
    return 0; // keep the compiler happy
}
